package animation;

import models.AnimatedModel;
import models.BoneInfo;
import models.Node;
import org.lwjgl.PointerBuffer;
import org.lwjgl.assimp.AIAnimation;
import org.lwjgl.assimp.AINodeAnim;
import models.skeleton.Line;
import models.skeleton.Skeleton;
import transforms.Mat4;
import transforms.Mat4Identity;

public class Animator {
    private Mat4[]finalTransforms;
    private Animation animation;
    private AnimatedModel model;
    private Mat4 globalInverseTransform;
    private Skeleton skeleton;

    public Animator(Animation animation, AnimatedModel model) {
        this.animation = animation;
        this.model = model;
        skeleton = new Skeleton();
        finalTransforms = new Mat4[model.getBoneInfoHashMap().size()];
        globalInverseTransform = new Mat4Identity();
    }

    public void updateAnimation(double timeInSeconds, boolean useOffset){
        if (animation != null){
            double ticksPerSecond = (animation.getTicksForSecond() != 0 ? animation.getTicksForSecond() : 25.0d);
            double timeInTicks = timeInSeconds * ticksPerSecond;
            double animationTime = (timeInTicks % animation.getDuration());
            calculateTransformMatrices(animation.getRootNode(),new Mat4Identity(), animationTime, useOffset);
        }
    }

    private void calculateTransformMatrices(Node node, Mat4 parentTransform, double animationTime, boolean useOffset) {
        String nodeName = node.getName();
        Mat4 nodeTransform = node.getTransformation();
        BoneInfo boneInfo = model.getBoneInfoHashMap().get(nodeName);
        AINodeAnim aiNodeAnim = findAnimNode(nodeName);

        if (boneInfo != null){
            Bone bone = new Bone(aiNodeAnim);
            bone.update(animationTime);
            nodeTransform = bone.getLocalTransform();
        }

        Mat4 globalTransform = parentTransform.mul(nodeTransform);

        if (boneInfo != null){
            Mat4 offset = model.getBoneInfoHashMap().get(nodeName).getOffsetMatrix();
            Mat4 finalBoneTransform;
            if (offset != null && useOffset)
                finalBoneTransform = globalInverseTransform.mul(globalTransform).mul(offset);
            else
                finalBoneTransform = globalInverseTransform.mul(globalTransform);


            finalTransforms[boneInfo.getId()] = finalBoneTransform;
        }
        //recursion for every childNode
        for (int i = 0; i<node.getChildrenCount(); i++){
            calculateTransformMatrices(node.getChildren().get(i), globalTransform, animationTime, useOffset);
        }
    }

    public Skeleton calculateSkelet(){
        return calculateSkelet(animation.getRootNode(), -1);
    }
    private Skeleton calculateSkelet(Node node, int parentId) {
        String nodeName = node.getName();
        BoneInfo boneInfo = model.getBoneInfoHashMap().get(nodeName);
        int childId = -1;

        if (boneInfo != null){
            childId = boneInfo.getId();
            if (parentId!=-1)
                skeleton.addNewLine(new Line(parentId,childId));
        }

        //recursion for every childNode
        for (int i = 0; i<node.getChildrenCount(); i++){
            calculateSkelet(node.getChildren().get(i), childId);
        }
        return skeleton;
    }

    public  AINodeAnim findAnimNode(String nodeName){
        AIAnimation aiAnimation = animation.getAiAnimation();
        PointerBuffer aiChannels = aiAnimation.mChannels();
        for (int i = 0; i < aiAnimation.mNumChannels(); i++){
            AINodeAnim aiNodeAnim = AINodeAnim.create(aiChannels.get(i));

            if (nodeName.equals(aiNodeAnim.mNodeName().dataString()))
                return aiNodeAnim;
        }
        return null;
    }

    public Mat4[] getFinalTransforms() {
        return finalTransforms;
    }

    public Skeleton getSkelet() {
        return skeleton;
    }
}
