package animation;

import animation.keys.KeyPosition;
import animation.keys.KeyRotation;
import animation.keys.KeyScale;
import org.lwjgl.assimp.*;
import transforms.*;

import java.util.ArrayList;
import java.util.List;

public class Bone {

    private List<KeyPosition> positions;
    private List<KeyScale> scales;
    private List<KeyRotation> rotations;

    private int numPositionKeys;
    private int numScalingKeys;
    private int numRotationKeys;

    private Mat4 localTransform;


    public Bone(AINodeAnim aiNodeAnim) {
        //initialize lists
        positions = new ArrayList<>();
        scales = new ArrayList<>();
        rotations = new ArrayList<>();

        numPositionKeys = aiNodeAnim.mNumPositionKeys();
        numScalingKeys = aiNodeAnim.mNumScalingKeys();
        numRotationKeys =  aiNodeAnim.mNumRotationKeys();

        //Setting position keys
        setPositions(aiNodeAnim);

        //Setting scaling keys
        setScales(aiNodeAnim);

        //Setting rotation keys
        setRotations(aiNodeAnim);
    }

    private void setPositions(AINodeAnim aiNodeAnim) {
        AIVectorKey.Buffer aiVectorKeys = aiNodeAnim.mPositionKeys();
        for (int i = 0; i< numPositionKeys; i++){
            AIVector3D positionVec =  aiVectorKeys.get(i).mValue();
            double timeStamp = aiVectorKeys.get(i).mTime();
            KeyPosition keyPosition = new KeyPosition();
            keyPosition.setPosition(new Vec3D(positionVec.x(), positionVec.y(), positionVec.z()));
            keyPosition.setTimeStamp(timeStamp);
            positions.add(keyPosition);
        }
    }

    private void setScales(AINodeAnim aiNodeAnim) {
        AIVectorKey.Buffer aiScaleKeys = aiNodeAnim.mScalingKeys();
        for (int i = 0; i< numScalingKeys; i++){
            AIVector3D scalingVec = aiScaleKeys.get(i).mValue();
            double timeStamp = aiScaleKeys.get(i).mTime();
            KeyScale keyScale = new KeyScale();
            keyScale.setScale(new Vec3D(scalingVec.x(), scalingVec.y(),scalingVec.z()));
            keyScale.setTimeStamp(timeStamp);
            scales.add(keyScale);
        }
    }

    private void setRotations(AINodeAnim aiNodeAnim) {
        AIQuatKey.Buffer rotationKeys = aiNodeAnim.mRotationKeys();
        for (int i = 0; i< numRotationKeys; i++){
            AIQuaternion rotationQuat = rotationKeys.get(i).mValue();
            double timeStamp = rotationKeys.get(i).mTime();
            KeyRotation keyRotation = new KeyRotation();
            keyRotation.setRotation(new Quat(rotationQuat.w(),rotationQuat.x(),rotationQuat.y(),rotationQuat.z()));
            keyRotation.setTimeStamp(timeStamp);
            rotations.add(keyRotation);
        }
    }

    public void update(double animationTime){
        // interpolate position
        Mat4 translation = interpolatePosition(animationTime);

        //interpolate scale
        Mat4 scale = interpolateScale(animationTime);

        //interpolate rotation
        Mat4 rotation = interpolateRotation(animationTime);

        localTransform = translation.mul(rotation).mul(scale);
    }

    private Mat4 interpolateRotation(double animationTime) {
        int indexStart = getRotationIndex(animationTime);
        int indexEnd = indexStart + 1;
        KeyRotation startRot = rotations.get(indexStart);
        KeyRotation endRot = rotations.get(indexEnd);

        double factor = getFactor(startRot.getTimeStamp(),endRot.getTimeStamp(),animationTime);
        Quat start = startRot.getRotation();
        Quat end = endRot.getRotation();

        AIQuaternion aiStart = AIQuaternion.create();
        aiStart.set((float) start.getR(),(float)start.getI(),(float)start.getJ(),(float) start.getK());

        AIQuaternion aieEnd = AIQuaternion.create();
        aieEnd.set((float) end.getR(),(float)end.getI(),(float)end.getJ(),(float) end.getK());

        AIQuaternion aiResult = AIQuaternion.create();
        Assimp.aiQuaternionInterpolate(aiResult,aiStart,aieEnd,(float)factor);
        Quat result = new Quat(aiResult.w(),aiResult.x(),aiResult.y(),aiResult.z());

        return new Mat4(result.toRotationMatrix()).transpose();
    }

    private Mat4 interpolatePosition(double animationTime) {
        int indexStart = getPositionIndex(animationTime);
        int indexEnd = indexStart + 1;
        KeyPosition startPos = positions.get(indexStart);
        KeyPosition endPos = positions.get(indexEnd);

        double factor = getFactor(startPos.getTimeStamp(),endPos.getTimeStamp(),animationTime);

        Vec3D start = startPos.getPosition();
        Vec3D end = endPos.getPosition();
        Vec3D delta = end.sub(start);
        Vec3D interpolatedPosition = start.add(delta.mul(factor));

        return new Mat4(interpolatedPosition.getX(),interpolatedPosition.getY(),interpolatedPosition.getZ());
    }

    private Mat4Scale interpolateScale(double animationTime){
        int indexStart = getScalingIndex(animationTime);
        int indexEnd = indexStart + 1;
        //temporary
        double factor = getFactor(scales.get(indexStart).getTimeStamp(),scales.get(indexEnd).getTimeStamp(),animationTime);;
        Vec3D start = scales.get(indexStart).getScale();
        Vec3D end = scales.get(indexEnd).getScale();

        Vec3D delta = end.sub(start);
        Vec3D finalScale = start.add(delta.mul(factor));

        return new Mat4Scale(finalScale);
    }

    private double getFactor(double lastKeyTime, double nextKeyTime, double currentAnimTime){
        double midWayLength = currentAnimTime  - lastKeyTime;
        double deltaTime = nextKeyTime - lastKeyTime;
        return  midWayLength / deltaTime;
    }
    private int getPositionIndex(double animationTime){
        for (int i = 0; i < numPositionKeys - 1; i++){
            if (animationTime < positions.get(i + 1).getTimeStamp())
                return i;
        }
        return 0;
    }

    private int getScalingIndex(double animationTime){
        for (int i = 0; i < numScalingKeys - 1; i++){
            if (animationTime < scales.get(i + 1).getTimeStamp())
                return i;
        }
        return 0;
    }

    private int getRotationIndex(double animationTime){
        for (int i = 0; i < numRotationKeys - 1; i++){
            if (animationTime < rotations.get(i + 1).getTimeStamp())
                return i;
        }
        return 0;
    }

    public Mat4 getLocalTransform() {
        return localTransform;
    }

}
