package assimp.loaders;

import models.AnimatedModel;
import models.BoneInfo;
import models.MeshData;
import models.Vertex;
import org.lwjgl.PointerBuffer;
import org.lwjgl.assimp.*;
import transforms.*;
import utils.MathUtill;

import java.util.*;

import static org.lwjgl.assimp.Assimp.*;
import static org.lwjgl.assimp.Assimp.aiImportFile;

public class MeshLoader {
    private static int boneCounter = 0;
    private static final LinkedHashMap<String, BoneInfo> boneInfoHashMap = new LinkedHashMap<>();
    private static AIScene aiScene;


    public static AnimatedModel loadAnimatedModel(String resourcePath) throws Exception {
        return loadAnimatedModel(resourcePath,
                Assimp.aiProcess_Triangulate |
                        aiProcess_LimitBoneWeights |
                        aiProcess_ValidateDataStructure |
                        aiProcess_MakeLeftHanded);
    }


    private static AnimatedModel loadAnimatedModel(String resourcePath, int flags) throws Exception {
        aiScene = aiImportFile(resourcePath, flags);
        if (aiScene == null) {
            throw new Exception("Error loading model");
        }

        int numMeshes = aiScene.mNumMeshes();
        PointerBuffer aiMeshes = aiScene.mMeshes();
        MeshData[] meshes = new MeshData[numMeshes];
        for (int i = 0; i < numMeshes; i++) {
            AIMesh aiMesh = AIMesh.create(Objects.requireNonNull(aiMeshes).get(i));
            MeshData mesh = processMesh(aiMesh);
            meshes[i] = mesh;
        }

        return new AnimatedModel(meshes, boneInfoHashMap);
    }

    public static MeshData processMesh(AIMesh aiMesh) {
        List<Vertex> vertices = new ArrayList<>();
        List<Integer> indices = new ArrayList<>();

        processIndices(aiMesh, indices);
        processVertices(aiMesh, vertices);

        getBoneInfoForVertices(vertices, aiMesh);

        return new MeshData(indices, vertices);
    }


    private static void processIndices(AIMesh mesh, List<Integer> indices) {
        AIFace.Buffer faces = mesh.mFaces();

        for (int i = 0; i < mesh.mNumFaces(); i++) {
            AIFace face = faces.get(i);
            for (int j = 0; j < face.mNumIndices(); j++) {
                indices.add(face.mIndices().get(j));
            }
        }
    }

    private static void processVertices(AIMesh aiMesh, List<Vertex> vertices) {
        AIVector3D.Buffer vectors = aiMesh.mVertices();
        AIVector3D.Buffer normals = aiMesh.mNormals();
        for (int i = 0; i < aiMesh.mNumVertices(); i++) {
            AIVector3D vector = vectors.get(i);
            AIVector3D normal = normals.get(i);

            Vertex vertex = new Vertex();

            setBoneInfoDefault(vertex);

            vertex.setPosition(new Vec3D(vector.x(), vector.y(), vector.z()));
            vertex.setNormal(new Vec3D(normal.x(),normal.y(),normal.z()));
            vertices.add(vertex);
        }
    }

    private static void setBoneInfoDefault(Vertex vertex) {
        for (int i = 0; i < vertex.getMAX_BONE_INFLUENCE(); i++) {
            vertex.getVertexBoneInfo().setBoneIds(i, -1);
            vertex.getVertexBoneInfo().setWeights(i, 0.0f);
        }
    }

    private static void getBoneInfoForVertices(List<Vertex> vertices, AIMesh aiMesh) {
        //going through each bone and putting it to hash map
        int boneId;
        for (int boneIndex = 0; boneIndex < aiMesh.mNumBones(); boneIndex++) {
            AIBone aiBone = AIBone.create(aiMesh.mBones().get(boneIndex));
            String boneName = aiBone.mName().dataString();
            if (!boneInfoHashMap.containsKey(boneName)) {
                BoneInfo newBoneInfo = new BoneInfo(boneCounter, MathUtill.toMat4Matrix(aiBone.mOffsetMatrix()));
                boneInfoHashMap.put(boneName, newBoneInfo);
                boneId = boneCounter;
                boneCounter++;
            } else {
                boneId = boneInfoHashMap.get(boneName).getId();
            }

            extractWeights(aiBone,vertices,boneId);
        }
    }

    private static void extractWeights(AIBone aiBone,List<Vertex> vertices,int boneId ){
        //going through vertex weights
        AIVertexWeight.Buffer aiVertexWeights = aiBone.mWeights();
        for (int weightIndex = 0; weightIndex < aiBone.mNumWeights(); weightIndex++) {
            AIVertexWeight vertexWeight = aiVertexWeights.get(weightIndex);
            int vertexId = vertexWeight.mVertexId();
            float weight = vertexWeight.mWeight();
            Vertex currentVertex = vertices.get(vertexId);
            if (currentVertex == null)
                System.err.println("current vertex is null");
            //setting boneInfo for current vertex
            if (vertexId <= vertices.size())
                setVertexBoneInfo(currentVertex, boneId, weight);
        }
    }
    private static void setVertexBoneInfo(Vertex vertex, int boneId, float weight) {
        for (int i = 0; i < vertex.getMAX_BONE_INFLUENCE(); i++) {
            if (vertex.getVertexBoneInfo().getSingleBoneId(i) < 0) {
                vertex.getVertexBoneInfo().setBoneIds(i, boneId);
                vertex.getVertexBoneInfo().setWeights(i, weight);
                break;
            }
        }
    }

    public static AIScene getAIScene() {
        return aiScene;
    }
}
